import os

import os_tornado
import pytest
import tornado

from tests.test_commands.cmd import run


def test_command_startproject(tmpdir):
    paths = """
    temp/manager.py
    temp/app/settings.py
    temp/app/__init__.py
    temp/app/extensions/__init__.py
    temp/app/extensions/example_extension.py
    temp/app/request_handlers/__init__.py
    temp/app/request_handlers/example_handler.py
    """
    run('startproject temp %s' % tmpdir.strpath)
    for p in paths.split():
        assert tmpdir.join(p.strip()).exists()


def test_command_list():
    env = os.environ.copy()
    env["OS_TORNADO_SETTINGS_MODULE"] = 'tests.test_commands.settings'
    result, _ = run('list', env=env)
    assert '[ext/req/all]' in result
    result, _ = run('list ext', env=env)
    assert 'TempExtension' in result
    result, _ = run('list req', env=env)
    assert 'TempHandler' in result
    result, _ = run('list all', env=env)
    assert 'TempExtension' in result
    assert 'TempHandler' in result


def test_command_version():
    result, _ = run('version')
    assert os_tornado.__version__ in result


def test_unkown_cmd():
    cmd = 'unknown'
    result, _ = run(cmd)
    assert 'Unknown command: %s' % cmd in result


def test_cmdline():
    result, _ = run()
    for cmd in ['list', 'runserver', 'startproject', 'version']:
        assert cmd in result


def get_unused_port():
    import socket
    from contextlib import closing

    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('localhost', 0))
        return s.getsockname()[1]


def test_runserver():
    port = get_unused_port()
    env = os.environ.copy()
    env["OS_TORNADO_SETTINGS_MODULE"] = 'tests.test_commands.settings'
    env["TEST_CMD_CALLBACK_STOP"] = 'TRUE'
    stdout, stderr = run('runserver -p %d' % port, env=env)
    flags = [
        '[LOAD] SUCC Temp tests.test_commands.temp_extension.TempExtension',
        '[LOAD] SUCC / tests.test_commands.temp_handler.TempHandler',
        'listen port %d' % port,
    ]
    for flag in flags:
        assert flag in stderr
    assert 'STOP SUCC' in stdout

    port = get_unused_port()
    stdout, stderr = run('runserver -s PORT=%d' % port, env=env)
    assert 'listen port %d' % port in stderr

    port1 = get_unused_port()
    stdout, stderr = run('runserver -s PORT=%d -p %d' %
                         (port, port1), env=env)
    assert 'listen port %d' % port1 in stderr



import pytest
import json
import urlparse
import tornado.ioloop
from os_tornado.decorators import jsonify


def test_not_request_handler_class():
    @jsonify
    class NotRequestHander(object):
        pass
    with pytest.raises(TypeError):
        NotRequestHander()


@jsonify
class GetOK(tornado.web.RequestHandler):
    def get(self):
        return {'status': 'ok'}


@jsonify
class Get404(tornado.web.RequestHandler):
    def get(self):
        raise tornado.web.HTTPError(404)


application = tornado.web.Application([
    (r"/get/ok", GetOK),
    (r"/get/404", Get404),
])


@pytest.mark.skip(reason="pytest-tornado don't support lastest version")
@pytest.fixture
def app():
    return application


@pytest.mark.skip(reason="pytest-tornado don't support lastest version")
@pytest.mark.gen_test
def test_return_dict(http_client, base_url):
    response = yield http_client.fetch(urlparse.urljoin(base_url, '/get/ok'))
    assert response.code == 200
    assert json.loads(response.body) == {'status': 'ok'}


@pytest.mark.skip(reason="pytest-tornado don't support lastest version")
@pytest.mark.gen_test
def test_not_found(http_client, base_url):
    response = yield http_client.fetch(urlparse.urljoin(base_url, '/get/404'), raise_error=False)
    assert response.code == 404
    assert json.loads(response.body)['status_code'] == 404



import json
from tornado.web import RequestHandler


class TempHandler(RequestHandler):
    def initialize(self, **kwargs):
        self.handler_settings = kwargs

    def get(self):
        self.write(json.dumps(self.handler_settings))



"""
Extension
"""
import collections
import logging
from os_tornado.utils.module_utils import load_class


class ExtensionManager(object):
    def __init__(self, manager):
        self._manager = manager
        self._extensions = collections.OrderedDict()
        self._logger = logging.getLogger('ExtensionManager')
        self._loaded = False

    def load(self):
        if self._loaded:
            return
        exts = self._manager.settings['EXTENSIONS']
        if not exts:
            return
        for ext_settings in exts:
            extension_cls = self._load_extension(ext_settings)
            if not extension_cls:
                continue

            name = extension_cls.name
            if 'name' not in ext_settings:
                self._logger.warn('[LOAD] no name, will use %s' % name)
            else:
                name = ext_settings['name']

            ext_settings.pop('extension_class')
            if name not in self._extensions:
                self._extensions[name] = extension_cls
                self._logger.info('[LOAD] SUCC %s %s.%s %s',
                                  name, extension_cls.__module__,
                                  extension_cls.__class__.__name__,
                                  str(ext_settings) if ext_settings else '')
            else:
                self._logger.warn(
                    '[LOAD] SKIP already exist: %s, %s', name, str(ext_settings))
        self._loaded = True

    def _load_extension(self, ext_settings):
        ext = None
        if 'extension_class' not in ext_settings:
            self._logger.error(
                '[LOAD] no extension_class, %s' % str(ext_settings))
            return ext

        try:
            extension_class_str = ext_settings['extension_class']
            _cls = load_class(extension_class_str, Extension)
            if _cls:
                ext = _cls(self._manager, ext_settings)
            else:
                self._logger.error('[LOAD] invalid, %s', extension_class_str)

        except Exception as e:
            self._logger.error('[LOAD] %s, %s' % (e, str(ext_settings)))
        return ext

    def iter_extensions(self):
        for name in self._extensions:
            yield self._extensions[name]

    def setup(self):
        for ext in self._extensions.values():
            try:
                ext.setup()
            except Exception as e:
                self._logger.error('[SETUP] %s %s' % (ext.name, e))

    def cleanup(self):
        for ext in reversed(self._extensions.values()):
            try:
                ext.cleanup()
            except Exception as e:
                self._logger.error('[CLEANUP] %s %s' % (ext.name, e))
        self._extensions.clear()

    def run(self):
        for ext in self._extensions.values():
            try:
                ext.run()
            except Exception as e:
                self._logger.error('[RUN] %s %s' % (ext.name, e))

    def get_extension(self, name):
        return self._extensions.get(name, None)


class Extension(object):
    def __init__(self, manager, ext_settings):
        self._manager = manager
        self._ext_settings = ext_settings
        self._name = ext_settings.get('name', self.__class__.__name__).strip()
        self._logger = logging.getLogger('Extension/%s' % self._name)

    @property
    def name(self):
        return self._name

    @property
    def ext_settings(self):
        return self._ext_settings

    @property
    def settings(self):
        return self._manager.settigns

    def run(self):
        pass

    def setup(self):
        pass

    def cleanup(self):
        pass



import traceback
import wrapt
from tornado.web import RequestHandler


def _jsonify_common_operations(method):
    def _exec(instance, *args, **kwargs):
        response = method(instance, *args, **kwargs)
        if response is not None:
            instance.write(response)
    return _exec


def _jsonify_write_error(method):
    def _exec(instance, status_code, **kwargs):
        if instance.settings.get("serve_traceback") and "exc_info" in kwargs:
            # in debug mode, try to send a traceback
            instance.set_header('Content-Type', 'text/plain')
            for line in traceback.format_exception(*kwargs["exc_info"]):
                instance.write(line)
            instance.finish()
        else:
            response = {'status_code': status_code}
            if instance._reason is not None:
                response['reason'] = instance._reason

            instance.finish(response)
    return _exec


_OPERATION_WRAPPER = dict.fromkeys(['head', 'get', 'post', 'delete', 'patch', 'put', 'options'],
                                   _jsonify_common_operations)
_OPERATION_WRAPPER['write_error'] = _jsonify_write_error

_WRAPPED_CLASSES = set([])


@wrapt.decorator
def jsonify(wrapped, instance, args, kwargs):
    if not issubclass(wrapped, RequestHandler):
        raise TypeError('%s is not subclass of tornado.web.RequestHandler'
                        % wrapped.__name__)
    if wrapped in _WRAPPED_CLASSES:
        return wrapped(*args, **kwargs)
    for operation in _OPERATION_WRAPPER:
        if hasattr(wrapped, operation):
            method = getattr(wrapped, operation)
            wrapped_method = _OPERATION_WRAPPER[operation](method)
            setattr(wrapped, operation, wrapped_method)
    _WRAPPED_CLASSES.add(wrapped)

    return wrapped(*args, **kwargs)
